from cython cimport nogil, view
from libcpp.map cimport map
from libcpp.string cimport string
from libcpp.unordered_map cimport unordered_map
from libcpp.utility cimport pair
from libcpp.vector cimport vector

from typing import Iterable

import numpy as np

from cpython.ref cimport PyObject

from .pyunicode cimport PyUnicodeSmartPtr


cdef class BytesLabelEncoder:

	cdef map[string, size_t] _labels
	cdef vector[string] _classes

	def __cinit__(self):
		self._labels = map[string, size_t]()
		self._classes = vector[string]()

	def __init__(self):
		pass

	def partial_fit(self, seq):
		cdef string item

		for item in seq:
			self._labels.insert(pair[string, size_t](item, self._labels.size()))

	def fit(self, seq):
		self.partial_fit(seq)

	def transform(self, seq):
		cdef string item
		cdef int i

		cdef size_t[::1] out = view.array(shape=(len(seq), ), itemsize=sizeof(size_t), format="Q")

		for i, item in enumerate(seq):
			out[i] = self._labels.at(item)

		return np.asarray(out)

	def finalize(self):
		cdef pair[string, size_t] item
		with nogil:
			self._classes.resize(self._labels.size())

			for item in self._labels:
				self._classes[item.second] = item.first

	def inverse_transform(self, size_t[::1] seq):
		cdef size_t i, item

		if self._classes.size() == 0 and self._labels.size() > 0:
			self.finalize()

		cdef object[::1] out = view.array(shape=(len(seq), ), itemsize=sizeof(PyObject *), format="O")

		for i, item in enumerate(seq):
			out[i] = self._classes[item]

		return np.asarray(out)

	@property
	def labels(self):
		return self._labels

	@property
	def classes(self):
		return self._classes

	def __getstate__(self):
		return self.labels, self.classes

	def __setstate__(self, state):
		cdef bytes k
		cdef int v
		for k, v in state[0].items():
			self._labels.insert(pair[string, size_t](k, v))

cdef class StringLabelEncoder:

	cdef unordered_map[PyUnicodeSmartPtr, size_t] _labels
	cdef vector[PyUnicodeSmartPtr] _classes

	def __cinit__(self):
		self._labels = unordered_map[PyUnicodeSmartPtr, size_t]()
		self._classes = vector[PyUnicodeSmartPtr]()

	def __init__(self):
		pass

	def partial_fit(self, seq: Iterable[str]) -> None:
		cdef object item

		for item in seq:
			if not isinstance(item, str):
				raise TypeError(f"expected str, {type(item).__name__} found")

			self._labels.insert(pair[PyUnicodeSmartPtr, size_t](PyUnicodeSmartPtr(<PyObject *>item), self._labels.size()))

	def fit(self, seq):
		self.partial_fit(seq)

	def transform(self, seq):
		cdef object item
		cdef int i

		cdef size_t[::1] out = view.array(shape=(len(seq), ), itemsize=sizeof(size_t), format="Q")

		for i, item in enumerate(seq):
			if not isinstance(item, str):
				raise TypeError(f"expected bytes, {type(item)} found")

			out[i] = self._labels.at(PyUnicodeSmartPtr(<PyObject *>item))

		return np.asarray(out)

	def finalize(self):
		cdef pair[PyUnicodeSmartPtr, size_t] item
		with nogil:
			self._classes.resize(self._labels.size())

			for item in self._labels:
				self._classes[item.second] = item.first

	def inverse_transform(self, size_t[::1] seq):
		cdef size_t item
		cdef size_t i

		if self._classes.size() == 0 and self._labels.size() > 0:
			self.finalize()

		cdef object[::1] out = view.array(shape=(len(seq), ), itemsize=sizeof(PyObject *), format="O")

		for i, item in enumerate(seq):
			out[i] = <object>self._classes[item].get()

		return np.asarray(out)

	@property
	def labels(self):
		cdef dict d = dict()
		cdef pair[PyUnicodeSmartPtr, size_t] item

		for item in self._labels:
			d[<object>item.first.get()] = item.second

		return d

	@property
	def classes(self):
		if self._classes.size() == 0:
			return np.array([])

		cdef object[::1] out = view.array(shape=(self._classes.size(), ), itemsize=sizeof(PyObject *), format="O")
		cdef size_t i

		for i in range(self._classes.size()):
			out[i] = <object>self._classes[i].get()

		return np.asarray(out)

	def __getstate__(self):
		return self.labels, self.classes

	def __setstate__(self, state):
		cdef str k
		cdef int v
		for k, v in state[0].items():
			self._labels.insert(pair[PyUnicodeSmartPtr, size_t](PyUnicodeSmartPtr(<PyObject *>k), v))



# distutils: language=c++

from cpython.ref cimport PyObject


cdef extern from "pyunicode.hpp":
	cdef cppclass PyUnicodeSmartPtr:

		PyUnicodeSmartPtr() except +
		PyUnicodeSmartPtr(PyObject *) except +
		PyObject *get() const



import pickle  # nosec B403
from unittest import TestCase

import numpy as np

from encoders.encoder import BytesLabelEncoder, StringLabelEncoder


class EncoderTest(TestCase):
    def test_bytes(self):
        stuff = [b"asd", b"qwe", b"zxc"] * 2

        le = BytesLabelEncoder()
        le.partial_fit(stuff)
        le.partial_fit(stuff)
        self.assertEqual(le.labels, {b"asd": 0, b"qwe": 1, b"zxc": 2})
        np.testing.assert_array_equal(le.transform(stuff), np.array([0, 1, 2, 0, 1, 2]))

    def test_bytes_finalize_empty(self):
        le = BytesLabelEncoder()
        le.finalize()
        np.testing.assert_array_equal(le.classes, np.array([]))

    def test_bytes_inv(self):
        stuff = [b"asd", b"qwe", b"zxc"] * 2

        le = BytesLabelEncoder()
        le.partial_fit(stuff)
        encoded = le.transform(stuff)
        le.finalize()
        np.testing.assert_array_equal(le.classes, np.array(stuff[:3]))
        result = le.inverse_transform(encoded)
        np.testing.assert_array_equal(result, np.array(stuff))

    def test_bytes_transform_fail(self):
        stuff = [b"asd", b"qwe", b"zxc"] * 2

        le = BytesLabelEncoder()
        le.finalize()
        with self.assertRaises(IndexError):
            le.transform(stuff)

    def test_bytes_typeerror(self):
        with self.assertRaises(TypeError):
            le = BytesLabelEncoder()
            le.partial_fit(["asd"])

    def test_bytes_pickle(self):
        stuff = [b"asd", b"qwe", b"zxc"] * 2

        le = BytesLabelEncoder()
        le.partial_fit(stuff)
        le2 = pickle.loads(pickle.dumps(le))  # nosec B301
        self.assertEqual(le.labels, le2.labels)

    def test_str(self):
        stuff = ["asü", "😀", "zxä"] * 2

        le = StringLabelEncoder()
        le.partial_fit(stuff)
        le.partial_fit(stuff)
        self.assertEqual(le.labels, {"asü": 0, "😀": 1, "zxä": 2})
        np.testing.assert_array_equal(le.transform(stuff), np.array([0, 1, 2, 0, 1, 2]))

    def test_str_finalize_empty(self):
        le = StringLabelEncoder()
        le.finalize()
        np.testing.assert_array_equal(le.classes, np.array([]))

    def test_str_inv(self):
        stuff = ["asü", "😀", "zxä"] * 2

        le = StringLabelEncoder()
        le.partial_fit(stuff)
        encoded = le.transform(stuff)
        le.finalize()
        np.testing.assert_array_equal(le.classes, np.array(stuff[:3]))
        result = le.inverse_transform(encoded)
        np.testing.assert_array_equal(result, np.array(stuff))

    def test_str_transform_fail(self):
        stuff = ["asü", "😀", "zxä"] * 2

        le = StringLabelEncoder()
        le.finalize()
        with self.assertRaises(IndexError):
            le.transform(stuff)

    def test_str_typeerror(self):
        with self.assertRaisesRegex(TypeError, "expected str, bytes found"):
            le = StringLabelEncoder()
            le.partial_fit([b"asd"])

    def test_str_pickle(self):
        stuff = ["asü", "😀", "zxä"] * 2

        le = StringLabelEncoder()
        le.partial_fit(stuff)
        le2 = pickle.loads(pickle.dumps(le))  # nosec B301
        self.assertEqual(le.labels, le2.labels)


if __name__ == "__main__":
    import unittest

    unittest.main()



import platform
import sys

import numpy as np
from Cython.Build import cythonize
from setuptools import Extension, setup

machine = platform.machine().lower()
x86 = ("x86_64", "amd64", "i386", "x86", "i686")

if sys.platform.startswith("linux"):
    if machine in x86:
        cflags = ["-std=c++14", "-O2", "-mavx2", "-fopenmp", "-ffast-math"]
    else:
        cflags = ["-std=c++14", "-O2", "-fopenmp", "-ffast-math"]
elif sys.platform == "win32":
    if machine in x86:
        cflags = ["/std:c++14", "/O2", "/arch:AVX2", "/openmp", "/fp:fast"]
    else:
        cflags = ["/std:c++14", "/O2", "/openmp", "/fp:fast"]
elif sys.platform == "darwin":
    if machine in x86:
        cflags = ["-std=c++14", "-O2", "-mavx2", "-ffast-math"]
    else:
        cflags = ["-std=c++14", "-O2", "-ffast-math"]
else:
    cflags = []

cy_extensions = [
    Extension(
        "encoders.cyfuncs",
        ["encoders/cyfuncs.pyx"],
        include_dirs=[np.get_include(), "encoders"],
        extra_compile_args=cflags,
        language="c++",
        define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")],
    ),
]

compiler_directives = {
    "binding": False,
    "boundscheck": False,
    "wraparound": False,
    "annotation_typing": True,
    "warn.undeclared": True,
    "warn.unused": True,
    "warn.unused_arg": True,
    "warn.unused_result": True,
}

with open("readme.md", encoding="utf-8") as fr:
    long_description = fr.read()

if __name__ == "__main__":
    setup(
        name="encoders",
        version="0.0.4",
        description="Faster than scikit encoders",
        author="Dobatymo",
        long_description=long_description,
        long_description_content_type="text/markdown; charset=UTF-8",
        url="https://github.com/Dobatymo/encoders",
        classifiers=[
            "Intended Audience :: Developers",
            "License :: OSI Approved :: ISC License (ISCL)",
            "Operating System :: OS Independent",
            "Programming Language :: Python :: 3",
        ],
        install_requires=[],
        extras_require={"test": ["numpy"], "bench": ["genutility", "numpy", "pandas", "scikit-learn"]},
        packages=["encoders"],
        package_data={"encoders": ["py.typed"]},
        ext_modules=cythonize(cy_extensions, language_level=3, compiler_directives=compiler_directives),
        python_requires=">=3.7",
        zip_safe=False,
    )
